# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from .file import FileIO, FileIOReadbackOptions, FileIOWriteOptions
from .image import AnyImageIO, ImageIO, ImageIOReadbackOptions, ImageIOWriteOptions
from .io import IO, IOReadbackOptions, IOWriteOptions
from .raw import AnyRawIO, RawIO, RawIOReadbackOptions, RawIOWriteOptions

__all__ = [
    "IO",
    "IOWriteOptions",
    "IOReadbackOptions",
    "FileIO",
    "FileIOWriteOptions",
    "FileIOReadbackOptions",
    "RawIO",
    "AnyRawIO",
    "RawIOWriteOptions",
    "RawIOReadbackOptions",
    "ImageIO",
    "AnyImageIO",
    "ImageIOWriteOptions",
    "ImageIOReadbackOptions",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Final, Literal, Optional, TypeVar, Union

import numpy as np

from no_vtf.core.image import Image, ImageChannels, ImageDataTypes, ImageDataTypesHDR
from no_vtf.typing import mypyc_attr

from .modifier import ImageModifier

FloatingPointNumBits = Literal[16, 32]

_Id_contra = TypeVar("_Id_contra", bound=ImageDataTypes, contravariant=True)

_Ic = TypeVar("_Ic", bound=ImageChannels)


@mypyc_attr(allow_interpreted_subclasses=True)
class FPPrecisionModifier(
    ImageModifier[_Id_contra, _Ic, Union[_Id_contra, ImageDataTypesHDR], _Ic]
):
    def __init__(
        self,
        *,
        min: Optional[FloatingPointNumBits] = None,  # noqa: A002
        max: Optional[FloatingPointNumBits] = None,  # noqa: A002
    ) -> None:
        if min is not None and max is not None and min > max:
            raise ValueError(f"Unordered precisions: {min = } <= {max = }")

        self._min: Final = min
        self._max: Final = max

    def __call__(
        self, image: Image[_Id_contra, _Ic]
    ) -> Image[Union[_Id_contra, ImageDataTypesHDR], _Ic]:
        if not np.issubdtype(image.dtype, np.floating):
            return image

        fp_bits = np.dtype(image.dtype).itemsize * 8

        if self._min is not None and fp_bits < self._min:
            dtype = np.dtype(f"float{self._min}")
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(data=data, dtype=dtype, channels=image.channels)

        if self._max is not None and fp_bits > self._max:
            dtype = np.dtype(f"float{self._max}")
            data = image.data.map(lambda data: data.astype(dtype))
            return Image(data=data, dtype=dtype, channels=image.channels)

        return image



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from collections.abc import Sequence
from typing import Final, Generic, TypeVar

from no_vtf.core.texture import Texture
from no_vtf.typing import mypyc_attr

_T_contra = TypeVar("_T_contra", bound=Texture, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureFilter(Generic[_T_contra]):
    @abstractmethod
    def __call__(self, textures: Sequence[_T_contra]) -> Sequence[_T_contra]: ...


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureCombinedFilter(TextureFilter[_T_contra]):
    def __init__(self, filters: Sequence[TextureFilter[_T_contra]]) -> None:
        self.filters: Final = filters

    def __call__(self, textures: Sequence[_T_contra]) -> Sequence[_T_contra]:
        for texture_filter in self.filters:
            textures = texture_filter(textures)

        return textures



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Final, Optional

from no_vtf.core.texture import Texture
from no_vtf.typing import mypyc_attr

from .filter import TextureFilter


@mypyc_attr(allow_interpreted_subclasses=True)
class ResolutionFilter(TextureFilter[Texture]):
    def __init__(
        self,
        *,
        min: Optional[int] = None,  # noqa: A002
        max: Optional[int] = None,  # noqa: A002
        closest_as_fallback: bool = False,
    ) -> None:
        if min is not None and max is not None and min > max:
            raise ValueError(f"Unordered resolutions: {min = } <= {max = }")

        self.closest_as_fallback: Final = closest_as_fallback

        self._min: Final = min
        self._max: Final = max

    def __call__(self, textures: Sequence[Texture]) -> Sequence[Texture]:
        if self._min is self._max is None:
            return textures

        def resolution_filter(texture: Texture) -> bool:
            if self._min is not None:
                if not all(
                    resolution >= self._min for resolution in (texture.width, texture.height)
                ):
                    return False

            if self._max is not None:
                if not all(
                    resolution <= self._max for resolution in (texture.width, texture.height)
                ):
                    return False

            return True

        exact_matches = list(filter(resolution_filter, textures))
        if exact_matches or not self.closest_as_fallback:
            return exact_matches

        assert self._min is not None or self._max is not None
        num_pixels = (self._min or self._max or 0) * (self._max or self._min or 0)

        close_matches = {
            abs(num_pixels - texture.width * texture.height): texture for texture in textures
        }
        close_matches = dict(sorted(close_matches.items()))

        closest_match = list(close_matches.values())[0:1]
        return closest_match



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import inspect
import pathlib

from collections.abc import Sequence
from typing import Literal, Optional

import click
import click_option_group

import no_vtf

from ._click import GroupedOption, HelpFormatter, OptionGroup, Slice, echo

_ImageDynamicRange = Literal["ldr", "hdr"]


def _show_credits(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    credits = """
    no_vtf - Valve Texture Format Converter
    Copyright (C) b5327157

    https://git.sr.ht/~b5327157/no_vtf/
    https://pypi.org/project/no-vtf/
    https://developer.valvesoftware.com/wiki/no_vtf

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, version 3 only.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
    FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with
    this program. If not, see <https://www.gnu.org/licenses/>.
    """

    echo(inspect.cleandoc(credits))
    ctx.exit()


def _show_version(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    echo(no_vtf.__version__)
    ctx.exit()


click.Context.formatter_class = HelpFormatter


@click.command(name="no_vtf", no_args_is_help=True)
@click.argument(
    "paths",
    metavar="[--] PATH...",
    type=click.Path(path_type=pathlib.Path, exists=True),
    required=True,
    nargs=-1,
)
@click_option_group.optgroup("Conversion mode", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--animate/--no-animate",
    cls=GroupedOption,
    help="Output animated image file (default) / output each frame individually",
    type=bool,
    default=True,
)
@click_option_group.optgroup.option(
    "--raw",
    cls=GroupedOption,
    help="Extract image data as-is (without decoding)",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup("\n  Extraction", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--mipmaps",
    "-m",
    cls=GroupedOption,
    help="Extract all mipmaps",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--low-res-img",
    cls=GroupedOption,
    help="Extract low resolution image",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--min-resolution",
    cls=GroupedOption,
    help="Minimum mipmap resolution",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--max-resolution",
    cls=GroupedOption,
    help="Maximum mipmap resolution",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--closest-resolution",
    cls=GroupedOption,
    help="Fallback to the closest resolution if no exact match",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--frames",
    cls=GroupedOption,
    help="Frames to extract",
    type=Slice(),
)
@click_option_group.optgroup.option(
    "--faces",
    cls=GroupedOption,
    help="Faces to extract",
    type=Slice(),
)
@click_option_group.optgroup.option(
    "--slices",
    cls=GroupedOption,
    help="Slices to extract",
    type=Slice(),
)
@click_option_group.optgroup(
    "\n  Image decoding (not used with --raw)",
    cls=OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--dynamic-range",
    cls=GroupedOption,
    help="Override LDR/HDR auto-detection",
    type=click.Choice(["ldr", "hdr"], case_sensitive=False),
)
@click_option_group.optgroup.option(
    "--overbright-factor",
    cls=GroupedOption,
    help="Multiplicative factor used for decoding compressed HDR textures",
    show_default=True,
    type=float,
    default=16.0,
)
@click_option_group.optgroup(
    "\n  Image postprocessing (not used with --raw)",
    cls=OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--hdr-to-ldr",
    cls=GroupedOption,
    help="Convert HDR from linear sRGB to sRGB and output as clipped LDR",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup.option(
    "--separate-channels",
    cls=GroupedOption,
    help="Output the RGB/L and A channels separately",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup(
    "\n  Image output (not used with --raw)",
    cls=OptionGroup,
)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--ldr-format",
    "-f",
    cls=GroupedOption,
    help="LDR output format",
    metavar="SINGLE[|MULTI]",
    show_default=True,
    type=str,
    default="tiff|apng",
)
@click_option_group.optgroup.option(
    "--hdr-format",
    "-F",
    cls=GroupedOption,
    help="HDR output format",
    metavar="SINGLE[|MULTI]",
    show_default=True,
    type=str,
    default="exr",
)
@click_option_group.optgroup.option(
    "--fps",
    cls=GroupedOption,
    help="Frame rate used for animated image files",
    show_default=True,
    type=int,
    default=5,
)
@click_option_group.optgroup.option(
    "--compress/--no-compress",
    cls=GroupedOption,
    help="Control lossless compression",
    type=bool,
    default=None,
)
@click_option_group.optgroup("\n  Read/write control", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "write",
    "--always-write/--no-write",
    cls=GroupedOption,
    help="Write images",
    type=bool,
    default=None,
)
@click_option_group.optgroup.option(
    "readback",
    "--readback/--no-readback",
    cls=GroupedOption,
    help="Readback images",
    type=bool,
    default=False,
)
@click_option_group.optgroup("\n  Output destination", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--output-dir",
    "-o",
    "output_directory",
    cls=GroupedOption,
    help="Output directory",
    metavar="PATH",
    type=click.Path(path_type=pathlib.Path, exists=True, file_okay=False, dir_okay=True),
)
@click_option_group.optgroup.option(
    "--output-file",
    "-O",
    cls=GroupedOption,
    help="Output file",
    metavar="PATH",
    type=click.Path(path_type=pathlib.Path, file_okay=True, dir_okay=False),
)
@click_option_group.optgroup("\n  Miscellaneous", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.option(
    "--num-workers",
    cls=GroupedOption,
    help="Number of workers for parallel conversion",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click_option_group.optgroup.option(
    "--no-progress",
    cls=GroupedOption,
    help="Do not show the progress bar",
    type=bool,
    is_flag=True,
)
@click_option_group.optgroup("\n  Info", cls=OptionGroup)  # type: ignore[misc]
@click_option_group.optgroup.help_option("--help", "-h", cls=GroupedOption)
@click_option_group.optgroup.option(
    "--version",
    cls=GroupedOption,
    help="Show the version and exit.",
    type=bool,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_version,
)
@click_option_group.optgroup.option(
    "--credits",
    cls=GroupedOption,
    help="Show the credits and exit.",
    type=bool,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_credits,
)
def main_command(
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path],
    output_file: Optional[pathlib.Path],
    ldr_format: str,
    hdr_format: str,
    dynamic_range: Optional[_ImageDynamicRange],
    mipmaps: bool,
    min_resolution: Optional[int],
    max_resolution: Optional[int],
    closest_resolution: bool,
    frames: Optional[slice],
    faces: Optional[slice],
    slices: Optional[slice],
    animate: bool,
    fps: int,
    separate_channels: bool,
    overbright_factor: float,
    hdr_to_ldr: bool,
    low_res_img: bool,
    compress: Optional[bool],
    raw: bool,
    write: Optional[bool],
    readback: bool,
    num_workers: Optional[int],
    no_progress: bool,
) -> None:
    """
    Convert Valve Texture Format files into standard image files.

    PATH can be either file, or directory (in which case it is recursively searched
    for .vtf files, symbolic links are not followed). Multiple paths may be provided.

    As the output path, it is possible to specify either file or directory.

    Specifying the output file is useful mostly for single-file conversions,
    with filters to ensure the output file will be written only once.

    If the output directory is not specified, images are output into the source directories
    (in-place conversion). Otherwise, directory tree for any found files will be reconstructed
    in the chosen directory.

    Output LDR/HDR format is selected by its common file name extension. It is recommended selecting
    one of the specifically supported image formats (PNG, APNG, TGA, TIFF, EXR). Other image formats
    have not been validated to work, but can be still selected. A secondary format specifically used
    to output animated image files can be selected after '|' (see default LDR format as an example).
    The "skip" format can be used to skip the write step entirely.

    For the specifically supported image formats, compression is configurable when saving the image.
    Lossless compression is enabled by default. Lossy compression is not used.

    The BGRA8888 format can store both LDR and compressed HDR images.
    The specific type is either auto-detected by looking at the input file name
    (roughly, if it contains "hdr" near the end), or can be set manually.

    It is possible to filter images to convert by min/max resolution (width & height),
    and by frames/faces/slices. The former supports exact or closest match. The latter
    supports selection by single index or via Python slicing:
    https://python-reference.readthedocs.io/en/latest/docs/brackets/slicing.html

    Face index mapping: right (0), left, back, front, up, down, sphere map (6).

    After applying filters, only the highest-resolution mipmap is converted by default.
    Alternatively, all mipmaps of the high-resolution image can be converted.

    Animated textures are converted into an animated multi-frame image file by default.
    Alternatively, they can also be converted into single-frame images with animation disabled.

    The RGB/L and A channels are packed into one file by default.
    When output separately, resulting file names will be suffixed with "_rgb", "_l" or "_a".

    By default, image files are only written if they do not exist already.
    Alternatively, they can be overwritten, or writing can be disabled entirely.

    Images can also be read back to verify they have been written properly.
    Readback will error if data to be written do not match what is in the file.

    Worker is spawned for each logical core to run the conversion in parallel.
    Number of workers can be overridden. If set to 1, conversion is sequential.
    Sequential conversion enables more verbose errors to be printed.

    Exit status: Zero if all went successfully, non-zero if there was an error.
    Upon a recoverable error, conversion will proceed with the next file.
    """

    main(
        paths=paths,
        output_directory=output_directory,
        output_file=output_file,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        dynamic_range=dynamic_range,
        mipmaps=mipmaps,
        min_resolution=min_resolution,
        max_resolution=max_resolution,
        closest_resolution=closest_resolution,
        frames=frames,
        faces=faces,
        slices=slices,
        animate=animate,
        fps=fps,
        separate_channels=separate_channels,
        overbright_factor=overbright_factor,
        hdr_to_ldr=hdr_to_ldr,
        low_res_img=low_res_img,
        compress=compress,
        raw=raw,
        write=write,
        readback=readback,
        num_workers=num_workers,
        no_progress=no_progress,
    )


def main(
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path] = None,
    output_file: Optional[pathlib.Path] = None,
    ldr_format: Optional[str] = None,
    hdr_format: Optional[str] = None,
    dynamic_range: Optional[_ImageDynamicRange] = None,
    mipmaps: Optional[bool] = None,
    min_resolution: Optional[int] = None,
    max_resolution: Optional[int] = None,
    closest_resolution: Optional[bool] = None,
    frames: Optional[slice] = None,
    faces: Optional[slice] = None,
    slices: Optional[slice] = None,
    animate: Optional[bool] = None,
    fps: Optional[int] = None,
    separate_channels: Optional[bool] = None,
    overbright_factor: Optional[float] = None,
    hdr_to_ldr: Optional[bool] = None,
    low_res_img: Optional[bool] = None,
    compress: Optional[bool] = None,
    raw: Optional[bool] = None,
    write: Optional[bool] = None,
    readback: Optional[bool] = None,
    num_workers: Optional[int] = None,
    no_progress: Optional[bool] = None,
) -> None:
    from ._main import main

    main(
        paths=paths,
        output_directory=output_directory,
        output_file=output_file,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        dynamic_range=dynamic_range,
        mipmaps=mipmaps,
        min_resolution=min_resolution,
        max_resolution=max_resolution,
        closest_resolution=closest_resolution,
        frames=frames,
        faces=faces,
        slices=slices,
        animate=animate,
        fps=fps,
        separate_channels=separate_channels,
        overbright_factor=overbright_factor,
        hdr_to_ldr=hdr_to_ldr,
        low_res_img=low_res_img,
        compress=compress,
        raw=raw,
        write=write,
        readback=readback,
        num_workers=num_workers,
        no_progress=no_progress,
    )



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

from typing import Literal

import numpy as np
import numpy.typing as npt
import PIL.BlpImagePlugin
import PIL.Image

from no_vtf.core.image import ImageWithRawData
from no_vtf.functional import Deferred


def decode_dxt1_rgb(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        rgba_uint8 = _decode_dxt_generic(
            encoded_image, logical_width, logical_height, n=1, pixel_format="DXT1"
        )
        rgb_uint8: npt.NDArray[np.uint8] = rgba_uint8[..., :3]
        return rgb_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_dxt1_rgba(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, n=1, pixel_format="DXT1"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt3(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, n=2, pixel_format="DXT3"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def decode_dxt5(
    encoded_image: bytes, logical_width: int, logical_height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    thunk = functools.partial(
        _decode_dxt_generic, encoded_image, logical_width, logical_height, n=3, pixel_format="DXT5"
    )
    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )


def _decode_dxt_generic(
    encoded_image: bytes, logical_width: int, logical_height: int, *, n: int, pixel_format: str
) -> npt.NDArray[np.uint8]:
    physical_width, physical_height = _dxt_physical_dimensions(logical_width, logical_height)

    # reference for "n" and "pixel_format": Pillow/src/PIL/DdsImagePlugin.py
    pil_image = PIL.Image.frombytes(  # pyright: ignore [reportUnknownMemberType]
        "RGBA", (physical_width, physical_height), encoded_image, "bcn", n, pixel_format
    )

    rgba_uint8: npt.NDArray[np.uint8] = np.array(pil_image)
    rgba_uint8 = rgba_uint8[:logical_height, :logical_width, :]
    return rgba_uint8


def _dxt_physical_dimensions(logical_width: int, logical_height: int) -> tuple[int, int]:
    physical_width = _dxt_physical_length(logical_width)
    physical_height = _dxt_physical_length(logical_height)
    return physical_width, physical_height


def _dxt_physical_length(logical_length: int) -> int:
    physical_length = (max(logical_length, 4) + 3) // 4 * 4
    return physical_length



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

from typing import Final, Optional

from no_vtf.core.image import AnyImageWithRawData, ImageDynamicRange
from no_vtf.core.image.decoder import ImageDecoder
from no_vtf.core.image.decoder.dxt import (
    decode_dxt1_rgb,
    decode_dxt1_rgba,
    decode_dxt3,
    decode_dxt5,
)
from no_vtf.core.image.decoder.raw import (
    decode_a_uint8,
    decode_abgr_uint8,
    decode_argb_uint8,
    decode_bgr_uint8,
    decode_bgra_uint4_le,
    decode_bgra_uint8,
    decode_l_uint8,
    decode_la_uint8,
    decode_rgb_uint8,
    decode_rgba_float16_le,
    decode_rgba_uint8,
    decode_uv_uint8,
)
from no_vtf.core.texture.decoder import TextureDecoder
from no_vtf.typing import mypyc_attr

from .generated.parser import VtfImageFormat as VtfParserImageFormat
from .image_decoder import (
    decode_bgr_uint8_bluescreen,
    decode_bgra_uint8_hdr,
    decode_rgb_uint8_bluescreen,
    decode_rgba_uint16_le_hdr,
)
from .texture import VtfTexture


@mypyc_attr(allow_interpreted_subclasses=True)
class VtfDecoder(TextureDecoder[VtfTexture, AnyImageWithRawData]):
    def __init__(
        self,
        *,
        dynamic_range: Optional[ImageDynamicRange] = None,
        overbright_factor: Optional[float] = None,
    ) -> None:
        self.dynamic_range: Final = dynamic_range
        self.overbright_factor: Final = overbright_factor

    def __call__(self, texture: VtfTexture) -> AnyImageWithRawData:
        decoder = self._get_decoder(texture)
        decoded_image = decoder(texture.image.image_data, texture.width, texture.height)
        return decoded_image

    def _get_decoder(self, texture: VtfTexture) -> ImageDecoder[AnyImageWithRawData]:
        image_format = texture.image.image_format

        dynamic_range = (
            self.dynamic_range if self.dynamic_range is not None else texture.dynamic_range
        )
        has_alpha_flag = texture.onebitalpha or texture.eightbitalpha

        decoder: Optional[ImageDecoder[AnyImageWithRawData]] = None
        match (image_format, dynamic_range, has_alpha_flag):
            case VtfParserImageFormat.rgba8888, _, _:
                decoder = decode_rgba_uint8
            case VtfParserImageFormat.abgr8888, _, _:
                decoder = decode_abgr_uint8
            case VtfParserImageFormat.rgb888, _, _:
                decoder = decode_rgb_uint8
            case VtfParserImageFormat.bgr888, _, _:
                decoder = decode_bgr_uint8
            case VtfParserImageFormat.i8, _, _:
                decoder = decode_l_uint8
            case VtfParserImageFormat.ia88, _, _:
                decoder = decode_la_uint8
            case VtfParserImageFormat.a8, _, _:
                decoder = decode_a_uint8
            case VtfParserImageFormat.rgb888_bluescreen, _, _:
                decoder = decode_rgb_uint8_bluescreen
            case VtfParserImageFormat.bgr888_bluescreen, _, _:
                decoder = decode_bgr_uint8_bluescreen
            case VtfParserImageFormat.argb8888, _, _:
                # VTFLib/VTFEdit, Gimp VTF Plugin, and possibly others, decode this format
                # differently because of mismatched channels (verified against VTF2TGA).
                decoder = decode_argb_uint8
            case VtfParserImageFormat.bgra8888, None, _:
                raise RuntimeError("Dynamic range is set neither on VtfTexture nor VtfDecoder")
            case VtfParserImageFormat.bgra8888, "ldr", _:
                decoder = decode_bgra_uint8
            case VtfParserImageFormat.bgra8888, "hdr", _:
                decoder = functools.partial(
                    decode_bgra_uint8_hdr, overbright_factor=self.overbright_factor
                )
            case VtfParserImageFormat.dxt1, _, False:
                decoder = decode_dxt1_rgb
            case VtfParserImageFormat.dxt1, _, True:
                decoder = decode_dxt1_rgba
            case VtfParserImageFormat.dxt3, _, _:
                decoder = decode_dxt3
            case VtfParserImageFormat.dxt5, _, _:
                decoder = decode_dxt5
            case VtfParserImageFormat.bgra4444, _, _:
                decoder = decode_bgra_uint4_le
            case VtfParserImageFormat.dxt1_onebitalpha, _, _:
                decoder = decode_dxt1_rgba
            case VtfParserImageFormat.uv88, _, _:
                decoder = decode_uv_uint8
            case VtfParserImageFormat.rgba16161616f, _, _:
                decoder = decode_rgba_float16_le
            case VtfParserImageFormat.rgba16161616, _, _:
                decoder = decode_rgba_uint16_le_hdr
            case _:
                pass

        if not decoder:
            raise NotImplementedError(f"Unsupported Valve texture format: {image_format.name}")

        return decoder



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Final, Generic, Literal, Optional, TypeVar, cast, overload

from typing_extensions import Self

# define every IO for IO.initialize()
import no_vtf.core.image.io  # noqa: F401  # pyright: ignore [reportUnusedImport]

from no_vtf.typing import mypyc_attr

_O = TypeVar("_O")
_T = TypeVar("_T")


@mypyc_attr(allow_interpreted_subclasses=True)
class _Options(Generic[_O]):
    def __init__(self, copy: Optional[_O] = None, /) -> None:
        if copy:
            self.merge(copy, non_none=False)

    def clone(self) -> Self:
        return type(self)(cast(_O, self))

    @abstractmethod
    def merge(self, other: _O, /, *, non_none: bool = True) -> Self:
        return self

    @overload
    @classmethod
    def rightmost(cls, first: _T, /, *, non_none: bool = True) -> _T: ...

    @overload
    @classmethod
    def rightmost(cls, first: _T, /, *rest: Optional[_T], non_none: Literal[True] = True) -> _T: ...

    @overload
    @classmethod
    def rightmost(
        cls, first: _T, /, *rest: Optional[_T], non_none: bool = True
    ) -> Optional[_T]: ...

    @classmethod
    def rightmost(cls, first: _T, /, *rest: Optional[_T], non_none: bool = True) -> Optional[_T]:
        old: Optional[_T] = first
        for new in rest:
            if new is not None or not non_none:
                old = new

        return old


@mypyc_attr(allow_interpreted_subclasses=True)
class IOWriteOptions(_Options["IOWriteOptions"]):
    def merge(self, other: IOWriteOptions, /, *, non_none: bool = True) -> Self:
        return super().merge(other, non_none=non_none)


@mypyc_attr(allow_interpreted_subclasses=True)
class IOReadbackOptions(_Options["IOReadbackOptions"]):
    def merge(self, other: IOReadbackOptions, /, *, non_none: bool = True) -> Self:
        return super().merge(other, non_none=non_none)


@mypyc_attr(allow_interpreted_subclasses=True)
class IO:
    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(
        self,
        *,
        write_defaults: IOWriteOptions = IOWriteOptions(),
        readback_defaults: IOReadbackOptions = IOReadbackOptions(),
        **kwargs: Any,
    ) -> None:
        if kwargs:
            raise TypeError(f"got unexpected keyword arguments: {[*kwargs.keys()]}")

        self._write_defaults: Final = write_defaults.clone()
        self._readback_defaults: Final = readback_defaults.clone()

    @property
    def write_defaults(self) -> IOWriteOptions:
        return self._write_defaults.clone()

    @property
    def readback_defaults(self) -> IOReadbackOptions:
        return self._readback_defaults.clone()



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Generic, TypeVar

from no_vtf.core.image import Image, ImageChannels, ImageDataTypes
from no_vtf.typing import mypyc_attr

_Id_co = TypeVar("_Id_co", bound=ImageDataTypes, covariant=True)
_Ic_co = TypeVar("_Ic_co", bound=ImageChannels, covariant=True)

_Id_contra = TypeVar("_Id_contra", bound=ImageDataTypes, contravariant=True)
_Ic_contra = TypeVar("_Ic_contra", bound=ImageChannels, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageModifier(Generic[_Id_contra, _Ic_contra, _Id_co, _Ic_co]):
    @abstractmethod
    def __call__(self, image: Image[_Id_contra, _Ic_contra]) -> Image[_Id_co, _Ic_co]: ...



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from .extractor import VtfExtractor
from .generated.parser import Vtf as VtfParser
from .namer import Vtf2TgaLikeNamer
from .texture import VtfTexture
from .texture_decoder import VtfDecoder

__all__ = [
    "VtfExtractor",
    "VtfParser",
    "Vtf2TgaLikeNamer",
    "VtfTexture",
    "VtfDecoder",
]



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
import re

from collections.abc import Sequence
from typing import IO, AnyStr, Final, Optional, TextIO, TypeVar, Union

import click
import click_option_group

_T = TypeVar("_T")


class HelpFormatter(click.HelpFormatter):
    def write_usage(self, prog: str, args: str = "", prefix: Optional[str] = None) -> None:
        prog = click.style(prog, fg=127, bold=True)
        args = click.style(args, bold=True)
        super().write_usage(prog, args, prefix)

    def write_heading(self, heading: str) -> None:
        heading = click.style(heading, underline=True)
        super().write_heading(heading)


class OptionGroup(click_option_group.OptionGroup):
    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if not help_record:
            return None

        name, help_ = help_record
        name = click.style(name, fg=172, bold=True)
        return name, help_


class GroupedOption(click_option_group.GroupedOption):
    _dim_pattern: Final[re.Pattern[str]] = re.compile(r"(?<=\s)\[[^\[\]\s]+\]", re.ASCII)

    def get_help_record(self, ctx: click.Context) -> Optional[tuple[str, str]]:
        help_record = super().get_help_record(ctx)
        if help_record is None:
            return None

        def dim_repl(match: re.Match[str]) -> str:
            return click.style(match.group(), dim=True)

        opts, opt_help = help_record
        opt_help = self._dim_pattern.sub(dim_repl, opt_help)
        return opts, opt_help


class Slice(click.ParamType):
    name = "slice"

    def get_metavar(self, param: click.Parameter) -> str:
        return "INDEX|[START]:[STOP][:STEP]"

    def convert(
        self,
        value: Union[str, slice],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> slice:
        if isinstance(value, slice):
            return value

        slice_indices = self._to_slice_indices(value, param, ctx)

        start = slice_indices[0]
        if len(slice_indices) == 1:
            if start is None:
                self.fail("Index is empty.", param, ctx)
            if start >= 0:
                return slice(start, start + 1)
            else:
                stop = start + 1 if start != -1 else None
                return slice(start, stop)

        stop = slice_indices[1]
        if len(slice_indices) == 2:
            return slice(start, stop)

        step = slice_indices[2]
        if len(slice_indices) == 3:
            if step == 0:
                self.fail("Slice step cannot be zero.", param, ctx)
            return slice(start, stop, step)

        self.fail(f"Too many values in {value!r}.", param, ctx)

    def _to_slice_indices(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> list[Optional[int]]:
        slice_indices: list[Optional[int]] = []
        for slice_index in value.split(":"):
            if not slice_index:
                slice_indices.append(None)
            else:
                try:
                    slice_indices.append(int(slice_index))
                except ValueError:
                    self.fail(f"{slice_index!r} is not a valid integer.", param, ctx)
        return slice_indices


def get_param_default(
    params: Sequence[click.core.Parameter], param_name: str, param_type: type[_T]
) -> _T:
    for param in params:
        if param.name == param_name:
            default = param.default
            if callable(default):
                default = default()

            if not isinstance(default, param_type):
                raise ValueError(f"{type(default) = } {param_type = }")

            return default

    raise ValueError(f"No such parameter: {param_name}")


def apply_param_default(
    params: Sequence[click.core.Parameter],
    param_name: str,
    param_type: type[_T],
    value: Optional[_T],
) -> _T:
    return value if value is not None else get_param_default(params, param_name, param_type)


def posix_tty_style(
    text: str,
    *,
    io: TextIO,
    fg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bg: Optional[Union[int, tuple[int, int, int], str]] = None,
    bold: Optional[bool] = None,
    dim: Optional[bool] = None,
    underline: Optional[bool] = None,
    overline: Optional[bool] = None,
    italic: Optional[bool] = None,
    blink: Optional[bool] = None,
    reverse: Optional[bool] = None,
    strikethrough: Optional[bool] = None,
    reset: bool = True,
) -> str:
    if os.name == "posix" and io.isatty():
        return click.style(
            text,
            fg=fg,
            bg=bg,
            bold=bold,
            dim=dim,
            underline=underline,
            overline=overline,
            italic=italic,
            blink=blink,
            reverse=reverse,
            strikethrough=strikethrough,
            reset=reset,
        )

    return text


def echo(
    message: str,
    file: Optional[IO[AnyStr]] = None,
    nl: bool = True,
    err: bool = False,
    color: Optional[bool] = None,
) -> None:
    message = click.style("", reset=True) + message
    click.echo(message, file, nl, err, color)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import collections
import functools
import pathlib
import re

from abc import abstractmethod
from collections import defaultdict
from collections.abc import Callable, Mapping, Sequence
from typing import Final, Generic, Literal, Optional, Protocol, TypeAlias, TypeVar, overload

from no_vtf.typing import mypyc_attr

from .image import AnyImage, AnyImageWithRawData, ImageChannels, ImageDataTypes, ImageDynamicRange
from .image.channel_separator import ChannelSeparator
from .image.io.file import FileIO, FileIOWriteOptions
from .image.io.image import AnyImageIO
from .image.io.raw import AnyRawIO
from .image.modifier import ImageModifier
from .texture import Texture
from .texture.decoder import TextureDecoder
from .texture.extractor import TextureExtractor
from .texture.filter import TextureFilter
from .texture.namer import TextureNamer

Quantity = Literal["single", "multi"]

ImageIOMapping: TypeAlias = Mapping[tuple[Quantity, ImageDynamicRange], AnyImageIO]

_T_contra = TypeVar("_T_contra", bound=Texture, contravariant=True)

_I = TypeVar("_I", bound=AnyImage)
_I_co = TypeVar("_I_co", bound=AnyImage, covariant=True)
_I2_co = TypeVar("_I2_co", bound=AnyImage, covariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class Receipt:
    def __init__(self) -> None:
        self.io_ready = False
        self.io_done = False
        self.output_written: Final[collections.Counter[pathlib.Path]] = collections.Counter()


@mypyc_attr(allow_interpreted_subclasses=True)
class Pipeline(Generic[_T_contra]):
    def __init__(
        self,
        *,
        input_extension_pattern: Optional[re.Pattern[str]] = None,
        animate: bool = False,
        separate_channels: bool = False,
        extractor: TextureExtractor[_T_contra],
        filter: Optional[TextureFilter[_T_contra]],  # noqa: A002
        decoder: TextureDecoder[_T_contra, AnyImageWithRawData],
        modifiers: Optional[
            Sequence[ImageModifier[ImageDataTypes, ImageChannels, ImageDataTypes, ImageChannels]]
        ] = None,
        namer: TextureNamer[_T_contra],
        image_io_write: Optional[ImageIOMapping] = None,
        image_io_readback: Optional[ImageIOMapping] = None,
        raw_io_write: Optional[AnyRawIO] = None,
        raw_io_readback: Optional[AnyRawIO] = None,
    ) -> None:
        self._input_extension_pattern: Final = input_extension_pattern

        self._animate: Final = animate
        self._separate_channels: Final = separate_channels

        self._extractor: Final = extractor
        self._filter: Final = filter
        self._decoder: Final = decoder
        self._namer: Final = namer

        self._image_io_write: Final = image_io_write or {}
        self._image_io_readback: Final = image_io_readback or {}

        self._raw_io_write: Final = raw_io_write
        self._raw_io_readback: Final = raw_io_readback

        self._apply_modifiers: Final = (
            functools.partial(_apply_modifiers, modifiers=modifiers) if modifiers else None
        )

    @overload
    def __call__(
        self,
        input_file: pathlib.Path,
        /,
        *,
        output_file: pathlib.Path,
        receipt: Optional[Receipt] = None,
    ) -> Receipt: ...

    @overload
    def __call__(
        self,
        input_file: pathlib.Path,
        /,
        *,
        output_directory: pathlib.Path,
        receipt: Optional[Receipt] = None,
    ) -> Receipt: ...

    def __call__(
        self,
        input_file: pathlib.Path,
        /,
        *,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
        receipt: Optional[Receipt] = None,
    ) -> Receipt:
        receipt = receipt or Receipt()

        items = self._get_items(input_file)

        if self._image_io_write or self._image_io_readback:
            self._process_items_image(items, receipt, output_file, output_directory)

        if self._raw_io_write or self._raw_io_readback:
            self._process_items_raw(items, receipt, output_file, output_directory)

        return receipt

    def _get_items(self, input_file: pathlib.Path) -> Sequence[_PipelineItem[AnyImageWithRawData]]:
        textures = self._extractor(input_file)

        if self._filter:
            textures = self._filter(textures)

        items: list[_PipelineItem[AnyImageWithRawData]] = []
        for texture in textures:
            image: AnyImageWithRawData = self._decoder(texture)

            input_name = input_file.name
            if self._input_extension_pattern:
                input_name = re.sub(self._input_extension_pattern, "", input_name)

            output_stem = self._namer(texture, input_name=input_name)

            item = _PipelineItem(sequence=[image], output_stem=output_stem)
            items.append(item)

        return items

    def _process_items_image(
        self,
        items: Sequence[_PipelineItem[AnyImage]],
        receipt: Receipt,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> Receipt:
        do_write_io = _DoImageIO(self._image_io_write, output_file, output_directory, receipt)
        do_readback_io = _DoImageIO(self._image_io_readback, output_file, output_directory, receipt)

        steps: Sequence[Optional[_PipelineStep[AnyImage]]] = [
            self._apply_modifiers,
            _separate_channels if self._separate_channels else None,
            _group_by_output_stem if self._animate else None,
            do_write_io.write_step,
            do_readback_io.readback_step,
        ]
        for step in steps:
            if step:
                items = step(items)

        return receipt

    def _process_items_raw(
        self,
        items: Sequence[_PipelineItem[AnyImageWithRawData]],
        receipt: Receipt,
        output_file: Optional[pathlib.Path] = None,
        output_directory: Optional[pathlib.Path] = None,
    ) -> Receipt:
        do_write_io = _DoRawIO(self._raw_io_write, output_file, output_directory, receipt)
        do_readback_io = _DoRawIO(self._raw_io_readback, output_file, output_directory, receipt)

        steps: Sequence[Optional[_PipelineStep[AnyImageWithRawData]]] = [
            _group_by_output_stem if self._animate else None,
            do_write_io.write_step,
            do_readback_io.readback_step,
        ]
        for step in steps:
            if step:
                items = step(items)

        return receipt


@mypyc_attr(allow_interpreted_subclasses=True)
class _PipelineItem(Generic[_I_co]):
    def __init__(self, *, sequence: Sequence[_I_co], output_stem: str) -> None:
        self.sequence: Final = sequence
        self.output_stem: Final = output_stem


class _PipelineStep(Protocol[_I]):
    def __call__(self, items: Sequence[_PipelineItem[_I]]) -> Sequence[_PipelineItem[_I]]: ...


def _apply_modifiers(
    items: Sequence[_PipelineItem[AnyImage]],
    *,
    modifiers: Sequence[
        ImageModifier[ImageDataTypes, ImageChannels, ImageDataTypes, ImageChannels]
    ],
) -> Sequence[_PipelineItem[AnyImage]]:
    def apply_modifier(
        image: AnyImage,
        modifier: ImageModifier[ImageDataTypes, ImageChannels, ImageDataTypes, ImageChannels],
    ) -> AnyImage:
        return modifier(image)

    return [
        _PipelineItem(
            sequence=[
                functools.reduce(apply_modifier, modifiers, image) for image in item.sequence
            ],
            output_stem=item.output_stem,
        )
        for item in items
    ]


def _separate_channels(
    items: Sequence[_PipelineItem[AnyImage]],
    *,
    channel_separator: ChannelSeparator = ChannelSeparator(),
) -> Sequence[_PipelineItem[AnyImage]]:
    new_items: list[_PipelineItem[AnyImage]] = []
    for item in items:
        new_sequence_by_channels: defaultdict[str, list[AnyImage]] = defaultdict(list)
        for image in item.sequence:
            for image_seperated in channel_separator(image):
                new_sequence_by_channels[image_seperated.channels].append(image_seperated)

        for channels, new_sequence in new_sequence_by_channels.items():
            new_output_stem = f"{item.output_stem}_{channels}"
            new_item = _PipelineItem(sequence=new_sequence, output_stem=new_output_stem)
            new_items.append(new_item)

    return new_items


def _group_by_output_stem(items: Sequence[_PipelineItem[_I]]) -> Sequence[_PipelineItem[_I]]:
    item_by_output_stem: dict[str, _PipelineItem[_I]] = {}
    for item in items:
        output_stem = item.output_stem
        if output_stem not in item_by_output_stem:
            item_by_output_stem[output_stem] = item
        else:
            old_item = item_by_output_stem[output_stem]
            new_sequence = list(old_item.sequence) + list(item.sequence)
            new_item = _PipelineItem(sequence=new_sequence, output_stem=output_stem)
            item_by_output_stem[output_stem] = new_item

    return list(item_by_output_stem.values())


@mypyc_attr(allow_interpreted_subclasses=True)
class _DoIO(Generic[_I]):
    def __init__(
        self,
        output_file: Optional[pathlib.Path],
        output_directory: Optional[pathlib.Path],
        receipt: Receipt,
    ) -> None:
        self.output_file: Final = output_file
        self.output_directory: Final = output_directory
        self.receipt: Final = receipt

    def write_step(self, items: Sequence[_PipelineItem[_I]]) -> Sequence[_PipelineItem[_I]]:
        return self._do_io_operations(items, self._write_io_operation)

    def readback_step(self, items: Sequence[_PipelineItem[_I]]) -> Sequence[_PipelineItem[_I]]:
        return self._do_io_operations(items, self._readback_io_operation)

    def _do_io_operations(
        self,
        items: Sequence[_PipelineItem[_I]],
        io_operation: Callable[[_PipelineItem[_I], FileIO[pathlib.Path, _I], pathlib.Path], bool],
    ) -> Sequence[_PipelineItem[_I]]:
        for item in items:
            io = self._get_io(item)
            output_file_path = self._get_output_file_path(item)
            if io and output_file_path:
                self.receipt.io_ready = True
                if io_operation(item, io, output_file_path):
                    self.receipt.io_done = True

        return items

    def _write_io_operation(
        self,
        item: _PipelineItem[_I],
        io: FileIO[pathlib.Path, _I],
        output_file_path: pathlib.Path,
    ) -> bool:
        options = FileIOWriteOptions()
        options.mkdir_parents = True

        if self.receipt.output_written[output_file_path]:
            options.overwrite = True

        if not io.write_file(output_file_path, *item.sequence, options=options):
            return False

        self.receipt.output_written[output_file_path] += 1
        return True

    def _readback_io_operation(
        self,
        item: _PipelineItem[_I],
        io: FileIO[pathlib.Path, _I],
        output_file_path: pathlib.Path,
    ) -> bool:
        return io.readback_file(output_file_path, *item.sequence)

    def _get_output_file_path(self, item: _PipelineItem[_I]) -> Optional[pathlib.Path]:
        if self.output_file:
            return self.output_file

        output_path_extension = self._get_output_path_extension(item)
        if not output_path_extension:
            return None

        output_name = f"{item.output_stem}.{output_path_extension}"

        assert self.output_directory is not None
        return self.output_directory / output_name

    @abstractmethod
    def _get_io(self, item: _PipelineItem[_I]) -> Optional[FileIO[pathlib.Path, _I]]: ...

    @abstractmethod
    def _get_output_path_extension(self, item: _PipelineItem[_I]) -> Optional[str]: ...


@mypyc_attr(allow_interpreted_subclasses=True)
class _DoRawIO(_DoIO[AnyImageWithRawData]):
    def __init__(
        self,
        raw_io: Optional[AnyRawIO],
        output_file: Optional[pathlib.Path],
        output_directory: Optional[pathlib.Path],
        receipt: Receipt,
    ) -> None:
        super().__init__(output_file, output_directory, receipt)
        self.raw_io: Final = raw_io

    def _get_io(self, item: _PipelineItem[_I]) -> Optional[AnyRawIO]:
        return self.raw_io

    def _get_output_path_extension(self, item: _PipelineItem[_I]) -> str:
        return "raw"


@mypyc_attr(allow_interpreted_subclasses=True)
class _DoImageIO(_DoIO[AnyImage]):
    def __init__(
        self,
        image_io_mapping: ImageIOMapping,
        output_file: Optional[pathlib.Path],
        output_directory: Optional[pathlib.Path],
        receipt: Receipt,
    ) -> None:
        super().__init__(output_file, output_directory, receipt)
        self.image_io_mapping: Final = image_io_mapping

    def _get_io(self, item: _PipelineItem[AnyImage]) -> Optional[AnyImageIO]:
        if not item.sequence:
            return None

        quantity: Quantity = "multi" if len(item.sequence) > 1 else "single"
        dynamic_range = item.sequence[0].dynamic_range

        for image in item.sequence:
            if image.dynamic_range != dynamic_range:
                raise ValueError("Dynamic range must be the same for all images")

        return self.image_io_mapping.get((quantity, dynamic_range))

    def _get_output_path_extension(self, item: _PipelineItem[_I]) -> Optional[str]:
        io = self._get_io(item)
        return io.image_io_format if io else None



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import contextlib
import functools
import pathlib
import re
import sys
import traceback

from collections.abc import Callable, Sequence
from typing import Final, Generic, Literal, Optional, TypedDict, TypeVar, cast, overload

import click

from ._alive_progress import alive_bar
from ._click import apply_param_default, echo, posix_tty_style
from .core.image import ImageChannels, ImageDataTypes, ImageDynamicRange
from .core.image.io.file import FileIOWriteOptions
from .core.image.io.image import AnyImageIO, ImageIO, ImageIOWriteOptions
from .core.image.io.raw import AnyRawIO, RawIOWriteOptions
from .core.image.modifier import ImageModifier
from .core.image.modifier.hdr_to_ldr_modifier import HdrToLdrModifier
from .core.pipeline import ImageIOMapping, Pipeline, Quantity, Receipt
from .core.texture import Texture
from .core.texture.filter import (
    FaceFilter,
    FrameFilter,
    MipmapFilter,
    ResolutionFilter,
    SliceFilter,
    TextureCombinedFilter,
    TextureFilter,
)
from .filesystem import InputPaths, OutputDirectories
from .task_runner import ParallelRunner, SequentialRunner, TaskRunner
from .vtf import Vtf2TgaLikeNamer, VtfDecoder, VtfExtractor, VtfTexture

_T = TypeVar("_T", bound=Texture)
_T_co = TypeVar("_T_co", bound=Texture, covariant=True)

_FORMAT_SKIP: Final[Literal["skip"]] = "skip"


class _PipelineIO(TypedDict, total=False):
    image_io_write: Optional[ImageIOMapping]
    image_io_readback: Optional[ImageIOMapping]
    raw_io_write: Optional[AnyRawIO]
    raw_io_readback: Optional[AnyRawIO]


def main(
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path] = None,
    output_file: Optional[pathlib.Path] = None,
    ldr_format: Optional[str] = None,
    hdr_format: Optional[str] = None,
    dynamic_range: Optional[ImageDynamicRange] = None,
    mipmaps: Optional[bool] = None,
    min_resolution: Optional[int] = None,
    max_resolution: Optional[int] = None,
    closest_resolution: Optional[bool] = None,
    frames: Optional[slice] = None,
    faces: Optional[slice] = None,
    slices: Optional[slice] = None,
    animate: Optional[bool] = None,
    fps: Optional[int] = None,
    separate_channels: Optional[bool] = None,
    overbright_factor: Optional[float] = None,
    hdr_to_ldr: Optional[bool] = None,
    low_res_img: Optional[bool] = None,
    compress: Optional[bool] = None,
    raw: Optional[bool] = None,
    write: Optional[bool] = None,
    readback: Optional[bool] = None,
    num_workers: Optional[int] = None,
    no_progress: Optional[bool] = None,
) -> None:
    from .main import main_command

    if output_file and output_directory:
        raise ValueError("Output file and directory is mutually exclusive")

    params = main_command.params
    ldr_format = apply_param_default(params, "ldr_format", str, ldr_format)
    hdr_format = apply_param_default(params, "hdr_format", str, hdr_format)
    mipmaps = apply_param_default(params, "mipmaps", bool, mipmaps)
    closest_resolution = apply_param_default(params, "closest_resolution", bool, closest_resolution)
    animate = apply_param_default(params, "animate", bool, animate)
    fps = apply_param_default(params, "fps", int, fps)
    separate_channels = apply_param_default(params, "separate_channels", bool, separate_channels)
    overbright_factor = apply_param_default(params, "overbright_factor", float, overbright_factor)
    hdr_to_ldr = apply_param_default(params, "hdr_to_ldr", bool, hdr_to_ldr)
    low_res_img = apply_param_default(params, "low_res_img", bool, low_res_img)
    raw = apply_param_default(params, "raw", bool, raw)
    readback = apply_param_default(params, "readback", bool, readback)
    no_progress = apply_param_default(params, "no_progress", bool, no_progress)

    vtf_extension_pattern = re.compile(r"\.vtf$", re.ASCII | re.IGNORECASE)

    texture_filters = _get_filters(
        mipmaps=mipmaps,
        min_resolution=min_resolution,
        max_resolution=max_resolution,
        closest_resolution=closest_resolution,
        frames=frames,
        faces=faces,
        slices=slices,
    )

    texture_extractor = VtfExtractor(low_res_img=low_res_img)
    texture_filter = TextureCombinedFilter(texture_filters)
    texture_decoder = VtfDecoder(dynamic_range=dynamic_range, overbright_factor=overbright_factor)
    texture_namer = Vtf2TgaLikeNamer(include_mipmap_level=mipmaps, include_frame=(not animate))

    modifiers: list[ImageModifier[ImageDataTypes, ImageChannels, ImageDataTypes, ImageChannels]] = (
        []
    )
    if hdr_to_ldr:
        modifiers.append(HdrToLdrModifier())

    pipeline_io, io_initializer = _create_io(
        write=write,
        readback=readback,
        raw=raw,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        compress=compress,
        fps=fps,
    )

    pipeline = Pipeline(
        input_extension_pattern=vtf_extension_pattern,
        animate=animate,
        separate_channels=separate_channels,
        extractor=texture_extractor,
        filter=texture_filter,
        decoder=texture_decoder,
        modifiers=modifiers,
        namer=texture_namer,
        **pipeline_io,
    )

    input_paths = InputPaths(paths)
    if input_paths.has_directories():
        _resolve_directories(input_paths, not no_progress)

    if output_file:
        tasks = _get_tasks(pipeline, input_paths, output_file=output_file)
    else:
        tasks = _get_tasks(pipeline, input_paths, output_directory=output_directory)

    task_runner: TaskRunner
    if (num_workers is None and len(tasks) > 1) or (num_workers and num_workers > 1):
        task_runner = ParallelRunner(max_workers=num_workers, initializer=io_initializer)
    else:
        task_runner = SequentialRunner()

    exit_status, io_ready, io_done = _process_tasks(task_runner, tasks, not no_progress)

    if write is None and not readback and tasks and exit_status == 0 and io_ready and not io_done:
        message = (
            click.style("Warning", fg="yellow")
            + ": No file was written. Did you mean to use the "
            + click.style("--always-write", bold=True)
            + " option?"
        )
        echo(message, file=sys.stderr)

    sys.exit(exit_status)


def _get_filters(
    *,
    mipmaps: bool,
    min_resolution: Optional[int],
    max_resolution: Optional[int],
    closest_resolution: bool,
    frames: Optional[slice],
    faces: Optional[slice],
    slices: Optional[slice],
) -> Sequence[TextureFilter[VtfTexture]]:
    texture_filters: list[TextureFilter[VtfTexture]] = []

    if frames:
        texture_filters.append(FrameFilter(frames=frames))
    if faces:
        texture_filters.append(FaceFilter(faces=faces))
    if slices:
        texture_filters.append(SliceFilter(slices=slices))
    if min_resolution is not None or max_resolution is not None:
        texture_filters.append(
            ResolutionFilter(
                min=min_resolution, max=max_resolution, closest_as_fallback=closest_resolution
            )
        )
    if not mipmaps:
        texture_filters.append(MipmapFilter(mipmap_levels=slice(-1, None), last="filtered"))

    return texture_filters


def _create_io(
    *,
    write: Optional[bool],
    readback: bool,
    raw: bool,
    ldr_format: str,
    hdr_format: str,
    compress: Optional[bool],
    fps: Optional[int],
) -> tuple[_PipelineIO, Optional[Callable[[], None]]]:
    base_write_defaults = FileIOWriteOptions()
    base_write_defaults.overwrite = write is True

    pipeline_io = _PipelineIO()
    io_initializer: Optional[Callable[[], None]] = None
    if not raw:
        ldr_format_split = ldr_format.split("|")
        hdr_format_split = hdr_format.split("|")

        non_skip_formats = [
            format_ for format_ in ldr_format_split + hdr_format_split if format_ != _FORMAT_SKIP
        ]
        io_initializer = functools.partial(ImageIO.initialize, non_skip_formats)
        io_initializer()

        ldr_format_single: Final = ldr_format_split[0]
        hdr_format_single: Final = hdr_format_split[0]

        ldr_format_multi: Final = (ldr_format_split[1:2] or ldr_format_split[0:1])[0]
        hdr_format_multi: Final = (hdr_format_split[1:2] or hdr_format_split[0:1])[0]

        image_io_write_defaults = ImageIOWriteOptions().merge(base_write_defaults)
        image_io_write_defaults.fps = fps
        image_io_write_defaults.compress = compress

        image_io: dict[tuple[Quantity, ImageDynamicRange], AnyImageIO] = {}
        if image_io_ldr_single := _create_image_io(ldr_format_single, image_io_write_defaults):
            image_io["single", "ldr"] = image_io_ldr_single
        if image_io_hdr_single := _create_image_io(hdr_format_single, image_io_write_defaults):
            image_io["single", "hdr"] = image_io_hdr_single
        if image_io_ldr_multi := _create_image_io(ldr_format_multi, image_io_write_defaults):
            image_io["multi", "ldr"] = image_io_ldr_multi
        if image_io_hdr_multi := _create_image_io(hdr_format_multi, image_io_write_defaults):
            image_io["multi", "hdr"] = image_io_hdr_multi

        pipeline_io["image_io_write"] = image_io if write is not False else None
        pipeline_io["image_io_readback"] = image_io if readback else None
    else:
        raw_io_write_defaults = RawIOWriteOptions().merge(base_write_defaults)
        raw_io = AnyRawIO(write_defaults=raw_io_write_defaults)

        pipeline_io["raw_io_write"] = raw_io if write is not False else None
        pipeline_io["raw_io_readback"] = raw_io if readback else None

    return pipeline_io, io_initializer


def _create_image_io(
    output_format: str,
    write_defaults: ImageIOWriteOptions,
    *,
    _unhandled_compression_formats: list[str] = [],
) -> Optional[AnyImageIO]:
    if output_format == _FORMAT_SKIP:
        return None

    image_io = AnyImageIO(image_io_format=output_format, write_defaults=write_defaults)

    if (
        write_defaults.compress is not None
        and ImageIOWriteOptions(image_io.write_defaults).compress is None
        and output_format not in _unhandled_compression_formats
    ):
        message = (
            click.style("Warning", fg="yellow")
            + ": Format "
            + click.style(image_io.image_io_format, bold=True)
            + " does not support compression control."
        )
        echo(message, file=sys.stderr)

        _unhandled_compression_formats.append(image_io.image_io_format)

    return image_io


def _resolve_directories(input_paths: InputPaths, show_progress: bool) -> None:
    progress_bar_manager = alive_bar(receipt=False) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        for file in input_paths.search_in_directories("*.[vV][tT][fF]", add_results=True):
            if progress_bar:
                progress_bar.text = posix_tty_style(file.name, io=sys.stderr, bold=True)
                progress_bar()
        input_paths.remove_directories()


@overload
def _get_tasks(
    pipeline: Pipeline[_T], input_paths: InputPaths, *, output_directory: Optional[pathlib.Path]
) -> Sequence[_Task[_T]]: ...


@overload
def _get_tasks(
    pipeline: Pipeline[_T], input_paths: InputPaths, *, output_file: pathlib.Path
) -> Sequence[_Task[_T]]: ...


def _get_tasks(
    pipeline: Pipeline[_T],
    input_paths: InputPaths,
    *,
    output_directory: Optional[pathlib.Path] = None,
    output_file: Optional[pathlib.Path] = None,
) -> Sequence[_Task[_T]]:
    output_directories = OutputDirectories(output_directory)

    tasks: list[_Task[_T]] = []
    for input_file, input_base_directory in input_paths:
        if output_file:
            task = _Task(pipeline=pipeline, input_file=input_file, output_file=output_file)
        else:
            output_directory = output_directories(input_file, input_base_directory)
            task = _Task(
                pipeline=pipeline, input_file=input_file, output_directory=output_directory
            )
        tasks.append(task)
    return tasks


def _process_tasks(
    task_runner: TaskRunner,
    tasks: Sequence[_Task[Texture]],
    show_progress: bool,
) -> tuple[int, bool, bool]:
    exit_status = 0
    io_ready = False
    io_done = False

    progress_bar_manager = alive_bar(len(tasks)) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        overwrite_warning_shown = False

        for task, result in task_runner(tasks):
            task = cast(_Task[Texture], task)
            if isinstance(result, Receipt):
                io_ready = result.io_ready or io_ready
                io_done = result.io_done or io_done

                if (
                    any(value > 1 for value in result.output_written.values())
                    and not overwrite_warning_shown
                ):
                    message = (
                        click.style("Warning", fg="yellow")
                        + ": During processing of "
                        + click.style(repr(task), bold=True)
                        + ", an output file was written to multiple times."
                        + " This can be avoided by using extraction filters."
                        + " This message will be shown only once."
                    )
                    echo(message, file=sys.stderr)

                    overwrite_warning_shown = True

                if progress_bar:
                    skipped = not result.io_done
                    progress_bar(skipped=skipped)
                    progress_bar.text = posix_tty_style(
                        str(task.input_file.name), io=sys.stderr, bold=True
                    )
            else:
                exit_status = 1

                exception: Exception = result
                formatted_exception = "".join(traceback.format_exception(exception))
                message = (
                    click.style("Error", fg="red")
                    + " while processing "
                    + click.style(repr(task), bold=True)
                    + f": {formatted_exception}"
                )
                echo(message, file=sys.stderr)

    return exit_status, io_ready, io_done


class _Task(Generic[_T_co]):
    @overload
    def __init__(
        self, *, pipeline: Pipeline[_T_co], input_file: pathlib.Path, output_directory: pathlib.Path
    ) -> None: ...

    @overload
    def __init__(
        self, *, pipeline: Pipeline[_T_co], input_file: pathlib.Path, output_file: pathlib.Path
    ) -> None: ...

    def __init__(
        self,
        *,
        pipeline: Pipeline[_T_co],
        input_file: pathlib.Path,
        output_directory: Optional[pathlib.Path] = None,
        output_file: Optional[pathlib.Path] = None,
    ) -> None:
        self.pipeline: Final = pipeline
        self.input_file: Final = input_file

        self._output_directory: Final = output_directory
        self._output_file: Final = output_file

    def __call__(self) -> Receipt:
        if self._output_file:
            return self.pipeline(self.input_file, output_file=self._output_file)
        else:
            assert self._output_directory
            return self.pipeline(self.input_file, output_directory=self._output_directory)

    def __str__(self) -> str:
        return f"{self.input_file}"

    def __repr__(self) -> str:
        return f"{str(self.input_file)!r}"



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Final

from no_vtf.core.texture import Texture
from no_vtf.typing import mypyc_attr

from .filter import TextureFilter


@mypyc_attr(allow_interpreted_subclasses=True)
class FaceFilter(TextureFilter[Texture]):
    def __init__(self, *, faces: slice, strict: bool = False) -> None:
        self.faces: Final = faces
        self.strict: Final = strict

    def __call__(self, textures: Sequence[Texture]) -> Sequence[Texture]:
        if not textures:
            return []

        face_counts = {texture.face.count for texture in textures if texture.face}
        if not face_counts:
            return [] if self.strict else textures

        if len(face_counts) != 1:
            raise ValueError("Face count must be the same for all filtered textures")

        indices = range(*self.faces.indices(next(iter(face_counts))))

        textures_filtered: list[Texture] = []
        for index in indices:
            textures_filtered.extend(
                texture
                for texture in textures
                if (texture.face and texture.face.index == index)
                or (not self.strict and not texture.face)
            )
        return textures_filtered



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Final, Literal, Union

from no_vtf.core.texture import Texture
from no_vtf.typing import mypyc_attr

from .filter import TextureFilter


@mypyc_attr(allow_interpreted_subclasses=True)
class MipmapFilter(TextureFilter[Texture]):
    def __init__(
        self,
        *,
        mipmap_levels: slice,
        last: Union[Literal["original"], Literal["filtered"]] = "original",
        strict: bool = False,
    ) -> None:
        self.mipmap_levels: Final = mipmap_levels
        self.last: Final = last
        self.strict: Final = strict

    def __call__(self, textures: Sequence[Texture]) -> Sequence[Texture]:
        if not textures:
            return []

        mipmap_counts = {texture.mipmap.count for texture in textures if texture.mipmap}
        if not mipmap_counts:
            return [] if self.strict else textures

        length: int
        if self.last == "original":
            if len(mipmap_counts) != 1:
                raise ValueError("Mipmap count must be the same for all filtered textures")

            length = next(iter(mipmap_counts))
        else:
            length = max(texture.mipmap.index for texture in textures if texture.mipmap) + 1

        indices = range(*self.mipmap_levels.indices(length))

        textures_filtered: list[Texture] = []
        for index in indices:
            textures_filtered.extend(
                texture
                for texture in textures
                if (texture.mipmap and texture.mipmap.index == index)
                or (not self.strict and not texture.mipmap)
            )
        return textures_filtered



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Literal

import numpy as np
import numpy.typing as npt

from no_vtf.core.image import ImageWithRawData
from no_vtf.functional import Deferred

from .ndarray import image_bytes_to_ndarray


def decode_rgb_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_rgba_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_argb_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (1, 2, 3, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_bgr_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgb")


def decode_bgra_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (2, 1, 0, 3), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_abgr_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (3, 2, 1, 0), np.uint8)
    )
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="rgba")


def decode_rgba_uint16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba_uint16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint16, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.uint16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.uint16), channels="rgba"
    )


def decode_rgba_float16_be(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, ">")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_rgba_float16_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.float16, Literal["rgba"]]:
    data = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1, 2, 3), np.float16, "<")
    )
    return ImageWithRawData(
        raw=encoded_image, data=data, dtype=np.dtype(np.float16), channels="rgba"
    )


def decode_l_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["l"]]:
    data = Deferred(lambda: image_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="l")


def decode_a_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["a"]]:
    data = Deferred(lambda: image_bytes_to_ndarray(encoded_image, width, height, (0,), np.uint8))
    return ImageWithRawData(raw=encoded_image, data=data, dtype=np.dtype(np.uint8), channels="a")


def decode_la_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["la"]]:
    la_uint8 = Deferred(
        lambda: image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
    )
    return ImageWithRawData(
        raw=encoded_image, data=la_uint8, dtype=np.dtype(np.uint8), channels="la"
    )


def decode_uv_uint8(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgb"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        rg_uint8 = image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)
        b_uint8: npt.NDArray[np.uint8] = np.zeros(rg_uint8.shape[:-1], dtype=np.uint8)
        rgb_uint8: npt.NDArray[np.uint8] = np.dstack((rg_uint8, b_uint8))
        return rgb_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgb"
    )


def decode_bgra_uint4_le(
    encoded_image: bytes, width: int, height: int
) -> ImageWithRawData[np.uint8, Literal["rgba"]]:
    def thunk() -> npt.NDArray[np.uint8]:
        bgra_uint4 = image_bytes_to_ndarray(encoded_image, width, height, (0, 1), np.uint8)

        br_uint8 = np.bitwise_and(np.left_shift(bgra_uint4, 4), 0xF0)
        ga_uint8 = np.bitwise_and(bgra_uint4, 0xF0)

        r_uint8 = br_uint8[..., [1]]
        g_uint8 = ga_uint8[..., [0]]
        b_uint8 = br_uint8[..., [0]]
        a_uint8 = ga_uint8[..., [1]]

        rgba_uint8: npt.NDArray[np.uint8] = np.dstack((r_uint8, g_uint8, b_uint8, a_uint8))
        return rgba_uint8

    return ImageWithRawData(
        raw=encoded_image, data=Deferred(thunk), dtype=np.dtype(np.uint8), channels="rgba"
    )



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

from typing import Final, Optional

from no_vtf.typing import mypyc_attr


@mypyc_attr(allow_interpreted_subclasses=True)
class Texture:
    def __init__(
        self,
        *,
        width: int,
        height: int,
        mipmap: Optional[TextureIndex],
        frame: Optional[TextureIndex],
        face: Optional[TextureIndex],
        slice: Optional[TextureIndex],  # noqa: A002
    ) -> None:
        self.width: Final = width
        self.height: Final = height

        self.mipmap: Final = mipmap
        self.frame: Final = frame
        self.face: Final = face
        self.slice: Final = slice


@mypyc_attr(allow_interpreted_subclasses=True)
class TextureIndex:
    def __init__(
        self,
        index: int,
        *,
        count: int,
    ) -> None:
        self.index: Final = index
        self.count: Final = count



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from collections.abc import Sequence
from typing import Final

from no_vtf.core.texture import Texture
from no_vtf.typing import mypyc_attr

from .filter import TextureFilter


@mypyc_attr(allow_interpreted_subclasses=True)
class FrameFilter(TextureFilter[Texture]):
    def __init__(self, *, frames: slice, strict: bool = False) -> None:
        self.frames: Final = frames
        self.strict: Final = strict

    def __call__(self, textures: Sequence[Texture]) -> Sequence[Texture]:
        if not textures:
            return []

        frame_counts = {texture.frame.count for texture in textures if texture.frame}
        if not frame_counts:
            return [] if self.strict else textures

        if len(frame_counts) != 1:
            raise ValueError("Frame count must be the same for all filtered textures")

        indices = range(*self.frames.indices(next(iter(frame_counts))))

        textures_filtered: list[Texture] = []
        for index in indices:
            textures_filtered.extend(
                texture
                for texture in textures
                if (texture.frame and texture.frame.index == index)
                or (not self.strict and not texture.frame)
            )
        return textures_filtered



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import functools
import itertools
import pathlib
import re

from collections.abc import Sequence
from typing import Any, ClassVar, Final, Literal, Optional, Protocol, TypeAlias, TypeVar, cast

import imageio.core.v3_plugin_api
import imageio.plugins.freeimage
import imageio.typing
import imageio.v3
import numpy as np
import numpy.typing as npt

from typing_extensions import Self

from no_vtf.core.image import AnyImage, ImageChannels, ImageDataTypes
from no_vtf.core.image.modifier.fp_precision_modifier import FPPrecisionModifier
from no_vtf.typing import mypyc_attr

from .file import FileIO, FileIOReadbackOptions, FileIOWriteOptions
from .io import IOReadbackOptions, IOWriteOptions

_IMAGE_IO_FORMAT_PATTERN: Final[re.Pattern[str]] = re.compile(
    r"[a-z0-9]+", re.ASCII | re.IGNORECASE
)

_T_contra = TypeVar("_T_contra", bound=pathlib.Path, contravariant=True)
_I_contra = TypeVar("_I_contra", bound=AnyImage, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageIOWriteOptions(FileIOWriteOptions):
    def __init__(self, copy: Optional[IOWriteOptions] = None, /) -> None:
        self.compress: Optional[bool] = None
        self.fps: Optional[int] = None

        super().__init__(copy)

    def merge(self, other: IOWriteOptions, /, *, non_none: bool = True) -> Self:
        if isinstance(other, ImageIOWriteOptions):
            self.compress = self.rightmost(self.compress, other.compress, non_none=non_none)
            self.fps = self.rightmost(self.fps, other.fps, non_none=non_none)

        return super().merge(other, non_none=non_none)


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageIOReadbackOptions(FileIOReadbackOptions):
    pass


@mypyc_attr(allow_interpreted_subclasses=True)
class ImageIO(FileIO[_T_contra, _I_contra]):
    @classmethod
    def initialize(
        cls, formats: Optional[Sequence[str]] = None, *, _recursive: bool = True
    ) -> None:
        super().initialize(_recursive=False)

        if not formats:
            _ImageIOBackend.initialize()
        else:
            for backend_format in map(str.lower, formats):
                match backend_format:
                    case "apng":
                        _ImageIOApngBackend.initialize()
                    case "exr":
                        _ImageIOExrBackend.initialize()
                    case "png":
                        _ImageIOPngBackend.initialize()
                    case "targa" | "tga":
                        _ImageIOTgaBackend.initialize()
                    case "tiff":
                        _ImageIOTiffBackend.initialize()
                    case _:
                        _ImageIOBackend.initialize()

        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize(formats)

    def __init__(self, *, image_io_format: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.image_io_format: Final = image_io_format

        if not _IMAGE_IO_FORMAT_PATTERN.fullmatch(self.image_io_format):
            raise ValueError(f"Invalid format: {self.image_io_format}")

        backend: _ImageIOBackend
        write_defaults = ImageIOWriteOptions(self.write_defaults)
        match self.image_io_format.lower():
            case "apng":
                backend = _ImageIOApngBackend()
            case "exr":
                write_defaults.fps = None
                backend = _ImageIOExrBackend()
            case "png":
                write_defaults.fps = None
                backend = _ImageIOPngBackend()
            case "targa" | "tga":
                write_defaults.fps = None
                backend = _ImageIOTgaBackend()
            case "tiff":
                write_defaults.fps = None
                backend = _ImageIOTiffBackend()
            case _:
                write_defaults.fps = None
                write_defaults.compress = None
                extension = f".{self.image_io_format}"
                backend = _ImageIOBackend(extension=extension)
        self._backend: Final = backend
        self._write_defaults.merge(write_defaults, non_none=False)

    def write_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOWriteOptions = IOWriteOptions(),
    ) -> bool:
        if not super().write_file(path, *images, options=options):
            return False

        merged_options = ImageIOWriteOptions(self.write_defaults).merge(options)
        self._backend.write(path, images, merged_options)
        return True

    def readback_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOReadbackOptions = IOReadbackOptions(),
    ) -> bool:
        if not super().readback_file(path, *images, options=options):
            return False

        merged_options = ImageIOReadbackOptions(self.readback_defaults).merge(options)
        self._backend.readback(path, images, merged_options)
        return True


AnyImageIO: TypeAlias = ImageIO[pathlib.Path, AnyImage]


@mypyc_attr(allow_interpreted_subclasses=True)
class _DummyImageIO(AnyImageIO):  # pyright: ignore [reportUnusedClass]
    pass


class _Opener(Protocol):
    def __call__(
        self,
        uri: imageio.typing.ImageResource,
        io_mode: Literal["r", "w"],
        *,
        extension: Optional[str] = None,
        format_hint: Optional[str] = None,
    ) -> imageio.core.v3_plugin_api.PluginV3: ...


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOBackend:
    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(self, *, extension: Optional[str] = None) -> None:
        self._extension: Final = extension

    def write(
        self,
        path: pathlib.Path,
        sequence: Sequence[AnyImage],
        options: ImageIOWriteOptions,
    ) -> None:
        opener = self._get_opener()
        with opener(path, "w", extension=self._extension) as image_resource:
            for image in sequence:
                kwargs = self._get_writer_kwargs(image, options)
                image = self._postprocess(image)
                data = self._get_data(image)

                image_resource.write(data, **kwargs)

    def readback(
        self,
        path: pathlib.Path,
        sequence: Sequence[AnyImage],
        options: ImageIOReadbackOptions,
    ) -> None:
        opener = self._get_opener()
        with opener(path, "r", extension=self._extension) as image_resource:
            image_resource_iter = (  # pyright: ignore [reportUnknownVariableType]
                image_resource.iter()  # pyright: ignore [reportUnknownMemberType]
            )
            for (
                image,
                read_data,  # pyright: ignore [reportUnknownVariableType]
            ) in itertools.zip_longest(  # pyright: ignore [reportUnknownVariableType]
                sequence, image_resource_iter  # pyright: ignore [reportUnknownArgumentType]
            ):
                if image is None or read_data is None:
                    raise ValueError(f"{path!r}: Number of frames differs from what is in the file")

                image = self._postprocess(image)
                data = self._get_data(image)

                if data.dtype != read_data.dtype:  # pyright: ignore [reportUnknownMemberType]
                    raise ValueError(f"{path!r}: Data type differs from what is in the file")

                if not self._compare_data(
                    data, read_data  # pyright: ignore [reportUnknownArgumentType]
                ):
                    raise ValueError(f"{path!r}: Data differs from what is in the file")

    def _get_opener(self) -> _Opener:
        return cast(_Opener, imageio.v3.imopen)  # pyright: ignore [reportUnknownMemberType]

    def _get_writer_kwargs(
        self, image: AnyImage, options: ImageIOWriteOptions
    ) -> dict[str, object]:
        return {}

    def _postprocess(self, image: AnyImage) -> AnyImage:
        return image

    def _get_data(self, image: AnyImage) -> npt.NDArray[np.generic]:
        data = image.data()

        # write luminance into three channels when alpha is present
        if image.channels == "la":
            l_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [0]]
            a_uint8: npt.NDArray[ImageDataTypes] = data[:, :, [1]]
            data = np.dstack((l_uint8, l_uint8, l_uint8, a_uint8))

        # remove last axis if its length is 1
        if data.shape[-1] == 1:
            data = data[..., 0]

        return data

    def _compare_data(
        self, data: npt.NDArray[np.generic], read_data: npt.NDArray[np.generic]
    ) -> bool:
        return np.array_equal(data, read_data)


_FP_FORCE_32_BITS: Final[FPPrecisionModifier[ImageDataTypes, ImageChannels]] = FPPrecisionModifier(
    min=32, max=32
)


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOPillowBackend(_ImageIOBackend):
    def __init__(self, *, extension: Optional[str] = None) -> None:
        super().__init__(extension=extension)

    def _get_opener(self) -> _Opener:
        # imopen is incorrectly typed
        return functools.partial(  # pyright: ignore [reportReturnType]
            imageio.v3.imopen, plugin="pillow"  # pyright: ignore [reportUnknownMemberType]
        )  # pyright: ignore [reportGeneralTypeIssues]


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOPngBackend(_ImageIOPillowBackend):
    def __init__(self, *, extension: str = ".png") -> None:
        super().__init__(extension=extension)

    def _get_writer_kwargs(
        self, image: AnyImage, options: ImageIOWriteOptions
    ) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image, options)
        compress = options.rightmost(True, options.compress)
        if not compress:
            kwargs["compress_level"] = 0
        return kwargs


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOApngBackend(_ImageIOPngBackend):
    def __init__(self) -> None:
        super().__init__(extension=".apng")

    def _get_writer_kwargs(
        self, image: AnyImage, options: ImageIOWriteOptions
    ) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image, options)
        if options.fps:
            kwargs["duration"] = 1000 / options.fps
        return kwargs


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOLegacyBackend(_ImageIOBackend):
    def __init__(
        self, *, imageio_format: Optional[str] = None, extension: Optional[str] = None
    ) -> None:
        super().__init__(extension=extension)
        self._imageio_format: Final = imageio_format

    def _get_opener(self) -> _Opener:
        # imopen is incorrectly typed
        return functools.partial(  # pyright: ignore [reportCallIssue]
            imageio.v3.imopen,  # pyright: ignore [reportUnknownMemberType]
            legacy_mode=True,
            plugin=self._imageio_format,
        )  # pyright: ignore [reportGeneralTypeIssues]


# IO_FLAGS is an implicit reexport
_FREEIMAGE_IO_FLAGS: Final[
    type[imageio.plugins.freeimage.IO_FLAGS]  # pyright: ignore [reportPrivateImportUsage]
] = imageio.plugins.freeimage.IO_FLAGS  # pyright: ignore [reportPrivateImportUsage]


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOFreeImageBackend(_ImageIOLegacyBackend):
    _freeimage_initialized: ClassVar[bool] = False

    @classmethod
    def initialize(cls, *, _recursive: bool = True) -> None:
        super().initialize(_recursive=False)

        if not _ImageIOFreeImageBackend._freeimage_initialized:
            # download() seems to be untyped because of implicit reexport
            imageio.plugins.freeimage.download()  # type: ignore[no-untyped-call]
            _ImageIOFreeImageBackend._freeimage_initialized = True

        if _recursive:
            for subclass in cls.__subclasses__():
                subclass.initialize()

    def __init__(self, *, imageio_format: str, extension: str) -> None:
        super().__init__(imageio_format=imageio_format, extension=extension)

        if not _ImageIOFreeImageBackend._freeimage_initialized:
            raise RuntimeError("ImageIO FreeImage backend was not initialized")

    def _get_writer_kwargs(
        self, image: AnyImage, options: ImageIOWriteOptions
    ) -> dict[str, object]:
        kwargs = super()._get_writer_kwargs(image, options)
        kwargs["flags"] = self._get_flags(image, options)
        return kwargs

    def _get_flags(self, image: AnyImage, options: ImageIOWriteOptions) -> int:
        return 0


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOExrBackend(_ImageIOFreeImageBackend):
    def __init__(self) -> None:
        super().__init__(imageio_format="EXR-FI", extension=".exr")

    def _get_flags(self, image: AnyImage, options: ImageIOWriteOptions) -> int:
        flags = super()._get_flags(image, options)
        compress = options.rightmost(True, options.compress)
        flags |= _FREEIMAGE_IO_FLAGS.EXR_ZIP if compress else _FREEIMAGE_IO_FLAGS.EXR_NONE
        if not np.issubdtype(image.data().dtype, np.float16):
            flags |= _FREEIMAGE_IO_FLAGS.EXR_FLOAT
        return flags

    def _postprocess(self, image: AnyImage) -> AnyImage:
        return _FP_FORCE_32_BITS(image)


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOTgaBackend(_ImageIOFreeImageBackend):
    def __init__(self) -> None:
        super().__init__(imageio_format="TARGA-FI", extension=".tga")

    def _get_flags(self, image: AnyImage, options: ImageIOWriteOptions) -> int:
        flags = super()._get_flags(image, options)
        compress = options.rightmost(True, options.compress)
        flags |= (
            _FREEIMAGE_IO_FLAGS.TARGA_SAVE_RLE if compress else _FREEIMAGE_IO_FLAGS.TARGA_DEFAULT
        )
        return flags


@mypyc_attr(allow_interpreted_subclasses=True)
class _ImageIOTiffBackend(_ImageIOFreeImageBackend):
    def __init__(self) -> None:
        super().__init__(imageio_format="TIFF-FI", extension=".tiff")

    def _get_flags(self, image: AnyImage, options: ImageIOWriteOptions) -> int:
        flags = super()._get_flags(image, options)
        compress = options.rightmost(True, options.compress)
        flags |= _FREEIMAGE_IO_FLAGS.TIFF_DEFAULT if compress else _FREEIMAGE_IO_FLAGS.TIFF_NONE
        return flags

    def _postprocess(self, image: AnyImage) -> AnyImage:
        return _FP_FORCE_32_BITS(image)



# SPDX-FileCopyrightText: b5327157 <b5327157@protonmail.com>
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import pathlib

from abc import abstractmethod
from typing import Generic, Optional, TypeVar

from typing_extensions import Self

from no_vtf.core.image import AnyImage
from no_vtf.typing import mypyc_attr

from .io import IO, IOReadbackOptions, IOWriteOptions

_T_contra = TypeVar("_T_contra", bound=pathlib.Path, contravariant=True)
_I_contra = TypeVar("_I_contra", bound=AnyImage, contravariant=True)


@mypyc_attr(allow_interpreted_subclasses=True)
class FileIOWriteOptions(IOWriteOptions):
    def __init__(self, copy: Optional[IOWriteOptions] = None, /) -> None:
        self.overwrite: Optional[bool] = None
        self.mkdir_parents: Optional[bool] = None

        super().__init__(copy)

    def merge(self, other: IOWriteOptions, /, *, non_none: bool = True) -> Self:
        if isinstance(other, FileIOWriteOptions):
            self.overwrite = self.rightmost(self.overwrite, other.overwrite, non_none=non_none)
            self.mkdir_parents = self.rightmost(
                self.mkdir_parents, other.mkdir_parents, non_none=non_none
            )

        return super().merge(other, non_none=non_none)


@mypyc_attr(allow_interpreted_subclasses=True)
class FileIOReadbackOptions(IOReadbackOptions):
    pass


@mypyc_attr(allow_interpreted_subclasses=True)
class FileIO(IO, Generic[_T_contra, _I_contra]):
    @abstractmethod
    def write_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOWriteOptions = IOWriteOptions(),
    ) -> bool:
        merged_options = FileIOWriteOptions(self.write_defaults).merge(options)
        overwrite = merged_options.rightmost(True, merged_options.overwrite)
        mkdir_parents = merged_options.rightmost(False, merged_options.mkdir_parents)

        if not overwrite and path.is_file():
            return False

        if mkdir_parents:
            path.parent.mkdir(parents=True, exist_ok=True)

        return True

    @abstractmethod
    def readback_file(
        self,
        path: _T_contra,
        /,
        *images: _I_contra,
        options: IOReadbackOptions = IOReadbackOptions(),
    ) -> bool:
        return True
